@@ -180,8 +180,8 @@ extern int   Parse __P((char *, char **));
 extern int   ParseEscape __P((struct user *, char *));
 extern void  DoScreen __P((char *, char **));
 extern int   IsNumColon __P((char *, int, char *, int));
-extern void  ShowWindows __P((void));
-extern char *AddWindows __P((char *, int, int));
+extern void  ShowWindows __P((int));
+extern char *AddWindows __P((char *, int, int, int));
 extern char *AddOtherUsers __P((char *, int, struct win *));
 extern int   WindowByNoN __P((char *));
 extern struct win *FindNiceWindow __P((struct win *, char *));
