@@ -22,10 +22,6 @@
  * $Id: window.h,v 1.11 1994/05/31 12:33:27 mlschroe Exp $ FAU
  */
 
-#ifndef MAXWIN
-# define MAXWIN	10
-#endif
-
 
 /* keep this in sync with the initialisations in window.c */
 struct NewWindow
